package de.akquinet.jbosscc.needle.ejb;

import de.akquinet.jbosscc.needle.db.InMemoryDatabaseUtil;
import de.akquinet.jbosscc.needle.inject.Injector;
import org.apache.commons.collections.map.CompositeMap;
import org.easymock.EasyMock;
import org.easymock.classextension.EasyMockSupport;
import org.junit.After;
import org.junit.Before;

import javax.annotation.Resource;
import javax.ejb.EJBContext;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.transaction.UserTransaction;
import java.lang.reflect.Field;
import java.lang.reflect.ParameterizedType;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * @author: bendt
 */
public class AbstractTestcase<T> extends EasyMockSupport {
    private final Class<T> clazz;
    protected Injector injector;
    private final Map<Object, Object> mocks = new HashMap<Object, Object>();
    private final Map<String, Object> resources = new HashMap<String, Object>();

    private UserTransaction mockTransaction;
    private EJBContext mockContext;
    private InMemoryDatabaseUtil databaseUtil;

    private T instance;
    protected Injector.InjectionObjectCreator lazyEmCreator;

    @Deprecated
    public AbstractTestcase(final Class<T> clazz) {
        this();
    }

     public AbstractTestcase() {
        this.clazz = (Class<T>) ((ParameterizedType) getClass().getGenericSuperclass()).getActualTypeArguments()[0];
    }

    protected Injector newInjector() {
        return new Injector();
    }

    private T createInstance() throws Exception {
        T instance = clazz.newInstance();

        return instance;
    }

    @Before
    public final void setupInstance() throws Exception {
        injector = newInjector();

        instance = createInstance();

        lazyEmCreator = new Injector.InjectionObjectCreator() {
            public Object getObjectToBeInjected() {
                return getDatabaseUtil().getEntityManager();
            }
        };

        injectEjbFields();
        injectResourceFields();

        injector.injectAllDeclaredFields(instance, PersistenceContext.class, lazyEmCreator);
    }

    private void injectResourceFields() {
        List<Field> resourceFields = injector.getAllResourceInjectedFields(instance);

        for (Field f : resourceFields) {
            Class type = f.getType();
            Resource res = f.getAnnotation(Resource.class);

            if (type == EJBContext.class) {
                mockContext = createNiceMock(EJBContext.class);
                mockTransaction = createNiceMock(UserTransaction.class);

                injector.injectResourceParameter(instance, res.mappedName(), mockContext);

                EasyMock.expect(mockContext.getUserTransaction()).andReturn(mockTransaction).anyTimes();
            } else {
                Object mock = createNiceMock(type);

                resources.put(res.mappedName(), mock);

                injector.injectResourceParameter(instance, res.mappedName(), mock);
            }
        }
    }

    private void injectEjbFields() {
        List<Field> ejbFields = injector.getAllEjbInjectedFields(instance);

        for (Field f : ejbFields) {
            Class type = f.getType();

            Object mock = createNiceMock(type);

            rememberMock(type, mock);

            injector.injectEjb(instance, type, mock);
        }
    }

    protected void rememberMock(Class type, Object mock) {
        mocks.put(type, mock);
    }

    @After
    public void cleanDatabase() {
        if (databaseUtil != null) {
            databaseUtil.deleteAll();
        }
    }

    protected InMemoryDatabaseUtil getDatabaseUtil() {
        if (databaseUtil == null) {
            databaseUtil = new InMemoryDatabaseUtil();
        }

        return databaseUtil;
    }

    public Injector getInjector() {
        return injector;
    }

    protected final EJBContext getMockContext() {
        return mockContext;
    }

    protected final UserTransaction getMockTransaction() {
        return mockTransaction;
    }

    protected final <X> X getResourceMock(final String mappedName) {
        return (X) resources.get(mappedName);
    }

    protected final <X> X getMock(final Class<X> key) {
        return (X) mocks.get(key);
    }

    protected T getObjectUnderTest() {
        return instance;
    }

    protected void forEachMock(final MockOperation operation) {
        Iterator<Map.Entry> i = new CompositeMap(mocks, resources).entrySet().iterator();

        while (i.hasNext()) {
            Map.Entry e = i.next();

            operation.doWithMock(e.getValue());
        }

        if (mockTransaction != null) {
            operation.doWithMock(mockTransaction);
        }

        if (mockContext != null) {
            operation.doWithMock(mockContext);
        }
    }

    protected interface MockOperation {
        void doWithMock(Object mock);
    }

    protected final <X> X resetToDefault(final Object mock) {
        EasyMock.resetToDefault(mock);
        return (X) mock;
    }

    protected final <X> X resetToStrict(final Object mock) {
        EasyMock.resetToStrict(mock);
        return (X) mock;
    }
}
